/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.adapter;

import org.objectweb.jorm.mapper.rdb.lib.RdbBlob;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapterException;
import org.objectweb.jorm.mapper.rdb.adapter.api.JoinedTable;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;

import java.io.ObjectStreamClass;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.sql.PreparedStatement;
import java.sql.Blob;
import java.sql.Connection;
import java.util.Date;
import java.util.List;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.io.Serializable;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Reader;
import java.io.InputStream;

/*
 * This class is a JDBC compliant implementation of the RdbAdapter interface.
 * That means a compliant JDBC driver should work with this implementation.
 * When a JDBC driver is not conform for some point, it is adviced to sub class
 * this implementation for redefining the functions which do no work.
 * 
 * @author	Sebastien Chassande-Barrioz
 */
public class BasicRdbAdapter implements RdbAdapter {

	private final static String SUBSTRING = "substring";
	private final static String CONCAT = "concat";
	private final static String FIRSTLOCATE = "position";
	private final static String INDEXEDLOCATE = "instr";

	/**
	 * Is the name of the RdbAdapter. The name is used, in particular, in the
	 * RdbAdapter factory for choosing an adapter. 
	 */
	protected String name;

	public BasicRdbAdapter() {
	}

	public BasicRdbAdapter(String name) {
		this.name = name;
	}

	public boolean supportBatchPreparedStatement() {
		return true;
	}

	public String getName() {
		return name;
	}

	/**
	 * This methods return the SQL type code linked to the java type
	 * specified in parameter.
	 * @return An int constant defined in the java.sql.Types class
	 */
	public int getSqlTypeCode(int typeCode) throws RdbAdapterException {
		return getSqlTypeCode(typeCode, null);
	}

	/**
	 * This methods return the SQL type code linked to the java type
	 * specified in parameter.
	 * @return An int constant defined in the java.sql.Types class
	 */
	public int getSqlTypeCode(int typeCode, String sqlColumnType) throws RdbAdapterException {
		switch (typeCode) {
		case TYPECODE_BOOLEAN:
		case TYPECODE_OBJBOOLEAN:
			return Types.BIT;
		case TYPECODE_CHAR:
		case TYPECODE_OBJCHAR:
			return Types.CHAR;
		case TYPECODE_BYTE:
		case TYPECODE_OBJBYTE:
			return Types.TINYINT;
		case TYPECODE_SHORT:
		case TYPECODE_OBJSHORT:
			return Types.SMALLINT;
		case TYPECODE_INT:
		case TYPECODE_OBJINT:
			return Types.INTEGER;
		case TYPECODE_LONG:
		case TYPECODE_OBJLONG:
			return Types.BIGINT;
		case TYPECODE_FLOAT:
		case TYPECODE_OBJFLOAT:
			return Types.REAL;
		case TYPECODE_DOUBLE:
		case TYPECODE_OBJDOUBLE:
			return Types.DOUBLE;
		case TYPECODE_STRING:
			return Types.VARCHAR;
		case TYPECODE_DATE:
			if (sqlColumnType == null) {
				return Types.TIMESTAMP;
			} else {
				String str = sqlColumnType.toUpperCase();
				if (str.startsWith("DATE")) {
					return Types.DATE;
				} else if (str.startsWith("TIMESTAMP")) {
					return Types.TIMESTAMP;
				} else if (str.startsWith("TIME")) {
					return Types.TIME;
				} else {
					return Types.TIMESTAMP;
				}
			}
		case TYPECODE_CHARARRAY:
			return Types.VARCHAR;
		case TYPECODE_BYTEARRAY:
		case TYPECODE_SERIALIZED:
			return Types.VARBINARY;
		case TYPECODE_BIGINTEGER:
			return Types.DECIMAL;
		case TYPECODE_BIGDECIMAL:
			return Types.DECIMAL;
		default:
			throw new RdbAdapterException("No sql type exist for the type code "
										 + typeCode);
		}
	}

	public String getSqlType(int typeCode, boolean usedInPK, int size, int scale) throws RdbAdapterException {
		switch (typeCode) {
		case TYPECODE_BOOLEAN:
		case TYPECODE_OBJBOOLEAN:
			return "BIT";
		case TYPECODE_CHAR:
		case TYPECODE_OBJCHAR:
			return "CHAR(1)";
		case TYPECODE_BYTE:
		case TYPECODE_OBJBYTE:
			return "TINYINT";
		case TYPECODE_SHORT:
		case TYPECODE_OBJSHORT:
			return "SMALLINT";
		case TYPECODE_INT:
		case TYPECODE_OBJINT:
			return "INTEGER";
		case TYPECODE_LONG:
		case TYPECODE_OBJLONG:
			return "BIGINT";
		case TYPECODE_FLOAT:
		case TYPECODE_OBJFLOAT:
			return "REAL";
		case TYPECODE_DOUBLE:
		case TYPECODE_OBJDOUBLE:
			return "DOUBLE";
		case TYPECODE_STRING:
		case TYPECODE_CHARARRAY:
			if (size != NOSIZE) {
				return "VARCHAR(" + size + ")";
			}
			return "VARCHAR";
		case TYPECODE_DATE:
			return "TIMESTAMP";
		case TYPECODE_BYTEARRAY:
		case TYPECODE_SERIALIZED:
			if (size != NOSIZE) {
				return "VARBINARY(" + size + ")";
			}
			return "VARBINARY";
		case TYPECODE_BIGINTEGER:
		case TYPECODE_BIGDECIMAL:
			if (size != NOSIZE) {
				if (scale != NOSIZE) {
					return "NUMERIC(" + size + ", " + scale + ")";
				} else {
					return "NUMERIC(" + size + ")";
				}
			} else {
				return "NUMERIC";
			}
		default:
			throw new RdbAdapterException("No sql type exist for the type code"
										 + typeCode);
		}
	}

	public void setNull(PreparedStatement ps, int idx, int sqlType) throws SQLException {
		ps.setNull(idx, sqlType);
	}

	/**
	 * This calculates the size of a result set. If this feature is not supported
	 * by a database or its jdbc driver, this method returns -1.
	 * @param rs the result set which the size must be returned
	 */
	public int fetchResultSetSize(ResultSet rs) throws SQLException {
		return -1;
	}

	public String getValueAsSQLString(Object value, int typeCode) {
		switch (typeCode) {
		case TYPECODE_BOOLEAN:
		case TYPECODE_OBJBOOLEAN:
			return (((Boolean) value).booleanValue() ? "1" : "0");

		case TYPECODE_OBJBYTE:
		case TYPECODE_OBJSHORT:
		case TYPECODE_OBJINT:
		case TYPECODE_OBJLONG:
		case TYPECODE_OBJFLOAT:
		case TYPECODE_OBJDOUBLE:
		case TYPECODE_BYTE:
		case TYPECODE_SHORT:
		case TYPECODE_INT:
		case TYPECODE_LONG:
		case TYPECODE_FLOAT:
		case TYPECODE_DOUBLE:
			return value.toString();

		case TYPECODE_OBJCHAR:
		case TYPECODE_STRING:
		case TYPECODE_DATE:
		case TYPECODE_BIGINTEGER:
		case TYPECODE_BIGDECIMAL:
			return "'" + value.toString() + "'";

		case TYPECODE_CHARARRAY:
		case TYPECODE_BYTEARRAY:
		case TYPECODE_SERIALIZED:
		default:
			return null;
		}
	}

	public String getValueAsSQLString(boolean value) {
		return (value ? "1" : "0");
	}

	public String getValueAsSQLString(byte value) {
		return "" + value;
	}

	public String getValueAsSQLString(char value) {
		return "'" + value + "'";
	}

	public String getValueAsSQLString(short value) {
		return "" + value;
	}

	public String getValueAsSQLString(int value) {
		return "" + value;
	}

	public String getValueAsSQLString(long value) {
		return "" + value;
	}

	public String getValueAsSQLString(float value) {
		return "" + value;
	}

	public String getValueAsSQLString(double value) {
		return "" + value;
	}

	public boolean getBoolean(ResultSet rs, int idx, boolean nullValue) throws SQLException {
		boolean res = rs.getBoolean(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Boolean getOboolean(ResultSet rs, int idx, Boolean nullValue) throws SQLException {
		boolean res = rs.getBoolean(idx);
		return (rs.wasNull() ? null : new Boolean(res));
	}

	public char getChar(ResultSet rs, int idx, char nullValue) throws SQLException {
		char res = rs.getString(idx).charAt(0);
		return (rs.wasNull() ? nullValue : res);
	}

	public Character getOchar(ResultSet rs, int idx, Character nullValue) throws SQLException {
		String s = rs.getString(idx);
		return (rs.wasNull() ? nullValue : new Character(s.charAt(0)));
	}

	public byte getByte(ResultSet rs, int idx, byte nullValue) throws SQLException {
		byte res = rs.getByte(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Byte getObyte(ResultSet rs, int idx, Byte nullValue) throws SQLException {
		byte res = rs.getByte(idx);
		return (rs.wasNull() ? null : new Byte(res));
	}

	public short getShort(ResultSet rs, int idx, short nullValue) throws SQLException {
		short res = rs.getShort(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Short getOshort(ResultSet rs, int idx, Short nullValue) throws SQLException {
		short res = rs.getShort(idx);
		return (rs.wasNull() ? nullValue : new Short(res));
	}

	public int getInt(ResultSet rs, int idx, int nullValue) throws SQLException {
		int res = rs.getInt(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Integer getOint(ResultSet rs, int idx, Integer nullValue) throws SQLException {
		int res = rs.getInt(idx);
		return (rs.wasNull() ? nullValue : new Integer(res));
	}

	public long getLong(ResultSet rs, int idx, long nullValue) throws SQLException {
		long res = rs.getLong(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Long getOlong(ResultSet rs, int idx, Long nullValue) throws SQLException {
		long res = rs.getLong(idx);
		return (rs.wasNull() ? nullValue : new Long(res));
	}

	public float getFloat(ResultSet rs, int idx, float nullValue) throws SQLException {
		float res = rs.getFloat(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Float getOfloat(ResultSet rs, int idx, Float nullValue) throws SQLException {
		float res = rs.getFloat(idx);
		return (rs.wasNull() ? null : new Float(res));
	}

	public double getDouble(ResultSet rs, int idx, double nullValue) throws SQLException {
		double res = rs.getDouble(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Double getOdouble(ResultSet rs, int idx, Double nullValue) throws SQLException {
		double res = rs.getDouble(idx);
		return (rs.wasNull() ? nullValue : new Double(res));
	}

	public String getString(ResultSet rs, int idx, String nullValue) throws SQLException {
		String res = rs.getString(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Date getDate(ResultSet rs, int idx, Date nullValue) throws SQLException {
		Date res = rs.getTimestamp(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Date getDate(ResultSet rs, int idx, String sqlColumnType, Date nullValue) throws SQLException {
		Date res = null;
		if (sqlColumnType == null || sqlColumnType.length() == 0) {
			res = rs.getTimestamp(idx);
		} else {
			String str = sqlColumnType.toUpperCase();
			if (str.startsWith("DATE")) {
				res = rs.getDate(idx);
			} else if (str.startsWith("TIMESTAMP")) {
				res = rs.getTimestamp(idx);
			} else if (str.startsWith("TIME")) {
				res = rs.getTime(idx);
			} else {
				res = rs.getTimestamp(idx);
			}
		}
		if (rs.wasNull()) {
			return nullValue;
		} else {
			return res;
		}
	}

	public char[] getCharArray(ResultSet rs, int idx, char[] nullValue) throws SQLException {
		Reader r = rs.getCharacterStream(idx);
		if (rs.wasNull()) {
			return nullValue;
		} else {
			try {
				return getCharArray(r);
			} catch (IOException e) {
				throw new SQLException(
						"Impossible to build a char[]: " + e.getMessage());
			}
		}
	}

	private static char[] getCharArray(java.io.Reader reader) throws java.io.IOException {
		int size = 0;
		long skipped = 0;
		while ((skipped = reader.skip(50)) == 50) {
			size += 50;
		}
		size += (int) skipped;
		char[] res = new char[size];
		reader.reset();
		reader.read(res);
		return res;
	}

	public byte[] getByteArray(ResultSet rs, int idx, byte[] nullValue) throws SQLException {
		byte[] res = rs.getBytes(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Serializable getSerialized(ResultSet rs, int idx,
            Serializable nullValue) throws SQLException, IOException,
            ClassNotFoundException {
        //		InputStream is = rs.getBinaryStream(idx);
        //		return (rs.wasNull() ? nullValue : (Serializable) new
        // ObjectInputStream(is).readObject());
        try {
            InputStream is = rs.getBinaryStream(idx);
            return (rs.wasNull() ? nullValue
                    : (Serializable) new ObjectInputStream(is).readObject());
        } catch (ClassNotFoundException e) {
            // Could not build serialized object:
            // Reason could be linked to class loaders hierarchy in an
            // application server
            class Specialized_OIS extends ObjectInputStream {
                Specialized_OIS(InputStream is) throws IOException {
                    super(is);
                }

                protected Class resolveClass(ObjectStreamClass osc)
                        throws IOException, ClassNotFoundException {
                    String n = osc.getName();
                    return Thread.currentThread().getContextClassLoader()
                            .loadClass(n);
                }
            }
            InputStream is = rs.getBinaryStream(idx);
            return (Serializable) new Specialized_OIS(is).readObject();
        }
    }
	
	public BigDecimal getBigDecimal(ResultSet rs, int idx, BigDecimal nullValue) throws SQLException {
		BigDecimal res = rs.getBigDecimal(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public BigInteger getBigInteger(ResultSet rs, int idx, BigInteger nullValue) throws SQLException {
		BigDecimal bd = getBigDecimal(rs, idx, null);
		if (bd == null) {
			return nullValue;
		} else {
			return bd.toBigInteger();
		}
	}

	public void setBoolean(PreparedStatement ps, int idx, boolean o)
			throws SQLException {
		ps.setBoolean(idx, o);
	}

	public void setOboolean(PreparedStatement ps, int idx, Boolean o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.BIT);
		else
			ps.setBoolean(idx, o.booleanValue());
	}

	public void setChar(PreparedStatement ps, int idx, char c)
			throws SQLException {
		ps.setString(idx, String.valueOf(c));
	}

	public void setOchar(PreparedStatement ps, int idx, Character o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.CHAR);
		else {
			ps.setString(idx, String.valueOf(o.charValue()));
		}
	}

	public void setByte(PreparedStatement ps, int idx, byte o)
			throws SQLException {
		ps.setByte(idx, o);
	}

	public void setObyte(PreparedStatement ps, int idx, Byte o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.SMALLINT);
		else
			ps.setByte(idx, o.byteValue());
	}

	public void setShort(PreparedStatement ps, int idx, short o)
			throws SQLException {
		ps.setShort(idx, o);
	}

	public void setOshort(PreparedStatement ps, int idx, Short o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.SMALLINT);
		else
			ps.setShort(idx, o.shortValue());
	}

	public void setInt(PreparedStatement ps, int idx, int o)
			throws SQLException {
		ps.setInt(idx, o);
	}

	public void setOint(PreparedStatement ps, int idx, Integer o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.INTEGER);
		else
			ps.setInt(idx, o.intValue());
	}

	public void setLong(PreparedStatement ps, int idx, long o)
			throws SQLException {
		ps.setLong(idx, o);
	}

	public void setOlong(PreparedStatement ps, int idx, Long o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.BIGINT);
		else
			ps.setLong(idx, o.longValue());
	}

	public void setFloat(PreparedStatement ps, int idx, float o)
			throws SQLException {
		ps.setFloat(idx, o);
	}

	public void setOfloat(PreparedStatement ps, int idx, Float o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.FLOAT);
		else
			ps.setFloat(idx, o.floatValue());
	}

	public void setDouble(PreparedStatement ps, int idx, double o)
			throws SQLException {
		ps.setDouble(idx, o);
	}

	public void setOdouble(PreparedStatement ps, int idx, Double o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.DOUBLE);
		else
			ps.setDouble(idx, o.doubleValue());
	}

	public void setString(PreparedStatement ps, int idx, String o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.VARCHAR);
		else
			ps.setString(idx, o);
	}

	public void setDate(PreparedStatement ps, int idx, Date o, String columnType)
			throws SQLException {
		if (columnType == null) {
//default is a Timestamp
			if (o == null)
				ps.setNull(idx, java.sql.Types.TIMESTAMP);
			else if (o instanceof java.sql.Timestamp) {
				ps.setTimestamp(idx, (java.sql.Timestamp) o);
			} else {
				ps.setTimestamp(idx, new java.sql.Timestamp(o.getTime()));
			}
		} else if (columnType.startsWith("DATE")) {
			if (o == null)
				ps.setNull(idx, java.sql.Types.DATE);
			else if (o instanceof java.sql.Date) {
				ps.setDate(idx, (java.sql.Date) o);
			} else {
				ps.setDate(idx, new java.sql.Date(o.getTime()));
			}
		} else if (columnType.startsWith("TIMESTAMP")) {
			if (o == null)
				ps.setNull(idx, java.sql.Types.TIMESTAMP);
			else if (o instanceof java.sql.Timestamp) {
				ps.setTimestamp(idx, (java.sql.Timestamp) o);
			} else {
				ps.setTimestamp(idx, new java.sql.Timestamp(o.getTime()));
			}
		} else if (columnType.startsWith("TIME")) {
			if (o == null)
				ps.setNull(idx, java.sql.Types.TIME);
			else if (o instanceof java.sql.Time) {
				ps.setTime(idx, (java.sql.Time) o);
			} else {
				ps.setTime(idx, new java.sql.Time(o.getTime()));
			}
		}
	}

	public void setCharArray(PreparedStatement ps, int idx, char[] o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.VARBINARY);
		else {
			ps.setCharacterStream(idx, new java.io.CharArrayReader(o), o.length);
		}
	}

	public void setByteArray(PreparedStatement ps, int idx, byte[] o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.VARBINARY);
		else
			ps.setBytes(idx, o);
	}

	public void setBigInteger(PreparedStatement ps, int idx, BigInteger o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.DECIMAL);
		else
			ps.setBigDecimal(idx, new BigDecimal(o));
	}

	public void setBigDecimal(PreparedStatement ps, int idx, BigDecimal o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.DECIMAL);
		else
			ps.setBigDecimal(idx, o);
	}

	public void setSerialized(PreparedStatement ps, int idx, Serializable o)
			throws SQLException, IOException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.VARBINARY);
		else {
			Blob blob = new RdbBlob(o);
			ps.setBinaryStream(idx, blob.getBinaryStream(), (int) blob.length());
		}
	}

	public String getSubstringName() {
		return SUBSTRING;
	}

	public String getConcatExpression(String op1, String op2) {
		return "{fn " + CONCAT + "(" + op1 + ", " + op2 + ")}";
	}

	public String getFirstLocateExpression(String substring, String instring)
			throws RdbAdapterException {
		return FIRSTLOCATE + "(" + substring + " in " + instring + ")";
	}

	public String getIndexedLocateExpression(String instring,
											 String substring,
											 String fromIndex)
			throws RdbAdapterException {
		return INDEXEDLOCATE + "(" + instring + ", " + substring + ", " +
				fromIndex + ")";
	}

	public String getLengthOperator() throws RdbAdapterException {
		return "length";
	}

	public String getQuery(String selectClause, List jts, String whereclause,
						   int rangeStart, int rangeSize) {
		StringBuffer sb = new StringBuffer();
		sb.append("SELECT ");
		modifySelectClauseWithRange(sb, rangeStart, rangeSize);
		sb.append(selectClause);

		sb.append(" FROM ");
		String sep = "";
		for (int i = 0; i < jts.size(); i++) {
			JoinedTable jt = (JoinedTable) jts.get(i);
			sb.append(sep);
			sep = ", ";
			getFromClause(jt, sb);
		}
		getWhereClause(jts, sb, whereclause, rangeStart, rangeSize);
		return sb.toString();
	}

	protected void getFromClause(JoinedTable jt, StringBuffer sb) {
		sb.append(jt.tableName);
		if (jt.alias != null && jt.alias.length() > 0) {
			writeTableAlias(jt.alias, sb);
		}
		for (int i = 0; i < jt.joinedTables.size(); i++) {
			getFromClause(jt,
						  (JoinedTable) jt.joinedTables.get(i),
						  (JoinedTable.Join) jt.joins.get(i),
						  sb);
		}
	}

	public void writeTableAlias(String alias, StringBuffer sb) {
		sb.append(" as ");
		sb.append(alias);
	}

    public void writeColumnAlias(String alias, StringBuffer sb) {
        sb.append(" as ");
        sb.append(alias);
    }

    public String getColumnAliasExpr(String aliasName) {
        return " as " + aliasName;
    }

    public String handleOrderBy(String query) {
		return query;
	}

	public String getCreateSequence(String seqName)
	    throws UnsupportedOperationException {
		return "CREATE SEQUENCE " + seqName;
	}

	public String getNextValInSequence(String seqName) {
		return "select nextVal('" + seqName + "')";
	}


    protected boolean existRelation(Connection connection, String relName, String[] relationTypes) throws SQLException {
        boolean existtable = false;
        ResultSet rs = null;
        
        if (relationTypes == null) {
            relationTypes = new String[]{"TABLE", "VIEW"};
        }
        
        try {
            rs = connection.getMetaData().getTables(null, null, relName, relationTypes);
            while (rs.next() && !existtable) {
                existtable = relName.equalsIgnoreCase(rs.getString(3));
            }
        } finally {
            if (rs != null) {
                rs.close();
            }
        }
        return existtable;
    }

    public boolean existTable(Connection connection, String tableName) throws SQLException {
		return existRelation(connection, tableName, new String[]{"TABLE"});
	}

    public boolean existView(Connection connection, String viewName) throws SQLException {
        return existRelation(connection, viewName, new String[]{"VIEW"});
    }

	public boolean existSequence(Connection connection, String seqName) throws SQLException {
		return existTable(connection, seqName);
	}


	protected void getFromClause(JoinedTable jtMain,
								 JoinedTable jt, JoinedTable.Join join,
								 StringBuffer sb) {
		sb.append(" left outer join ");
		sb.append(jt.tableName);
		if (jt.alias != null && jt.alias.length() > 0) {
			writeTableAlias(jt.alias, sb);
		}
		sb.append(" on ");
		String sep = "";
		String main = (jtMain.alias != null && jtMain.alias.length() > 0
				? jtMain.alias
				: jtMain.tableName );
		String ext = (jt.alias != null && jt.alias.length() > 0
				? jt.alias
				: jt.tableName );
		for (int i = 0; i < join.colT1.size(); i++) {
			sb.append(sep);
			sep = " AND ";
			sb.append(main);
			sb.append(".");
			sb.append(join.colT1.get(i));
			sb.append(" = ");
			sb.append(ext);
			sb.append(".");
			sb.append(join.colT2.get(i));
		}
		for (int i = 0; i < jt.joinedTables.size(); i++) {
			getFromClause(jt,
						  (JoinedTable) jt.joinedTables.get(i),
						  (JoinedTable.Join) jt.joins.get(i),
						  sb);
		}
	}

	protected void getWhereClause(List jts,
								  StringBuffer sb,
								  String whereclause,
								  int rangeStart,
								  int rangeSize) {
		if (whereclause != null && whereclause.trim().length() > 0) {
			sb.append(" WHERE ");
			sb.append(whereclause);
			modifyWhereClauseWithRange(sb, whereclause, rangeStart, rangeSize);
		}
	}

	public static char[] getCharArray(Object o) throws IOException, RdbAdapterException {
		if (o instanceof java.io.Reader) {
			java.io.Reader reader = (java.io.Reader) o;
			int size = 0;
			long skipped = 0;
			while ((skipped = reader.skip(50)) == 50) {
				size += 50;
			}
			size += (int) skipped;
			char[] res = new char[size];
			reader.reset();
			reader.read(res);
			return res;
		} else if (o instanceof String) {
			return ((String) o).toCharArray();
		} else {
			throw new RdbAdapterException("Unknown type to convert to char[]");
		}
	}

	public String getSubqueryAlias() {
		return "";
	}

	/**
	 * CREATE [ TEMPORARY | TEMP ] SEQUENCE name [ INCREMENT [ BY ] increment ]
    [ MINVALUE minvalue | NO MINVALUE ] [ MAXVALUE maxvalue | NO MAXVALUE ]
    [ START [ WITH ] start ] [ CACHE cache ] [ [ NO ] CYCLE ]
	 * @param seqName
	 * @param startid
	 * @param inc
	 */
	public String getCreateSequence(String seqName,
	                                Integer startid,
	                                Integer inc,
	                                Integer cache)
	        throws UnsupportedOperationException {
		StringBuffer sb = new StringBuffer("CREATE SEQUENCE ");
		sb.append(seqName);
		if (inc != null) {
			sb.append(" ");
			sb.append(getINCREMENTkeyWord());
			sb.append(" ");
			sb.append(inc);
		}
		if (startid != null) {
			sb.append(" ");
			sb.append(getSTARTkeyWord());
			sb.append(" ");
			sb.append(startid);
		}
		if (cache != null) {
			sb.append(" ");
			sb.append(getSTARTkeyWord());
			sb.append(" ");
			sb.append(cache);
		}
		return sb.toString();
	}

	protected String getINCREMENTkeyWord() {
		return "INCREMENT BY";
	}
	protected String getSTARTkeyWord() {
		return "START WITH";
	}
	protected String getCACHEkeyWord() {
		return "CACHE";
	}


	protected void modifySelectClauseWithRange(StringBuffer sb,
											  int rangeStart,
											  int rangeSize) {
		//TODO add log warning
	}

	protected void modifyWhereClauseWithRange(StringBuffer sb,
												String where,
												int rangeStart,
												int rangeSize) {
		//TODO add log warning
	}


}
