/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.adapter;

import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapterException;

import java.util.Map;
import java.util.HashMap;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.ResultSet;
import java.sql.Statement;

/**
 * @author S.Chassande-Barrioz
 */
public class OracleAdapter extends BasicRdbAdapter {

	private final static String PROP_BASE = "org.objectweb.jorm.adapter.oracle.";
	private final static String ORACLESUBSTRING = "substr";
	private final static String ORACLEFIRSTLOCATE = "instr";

	public final static String[][] PROPS = {
		{PROP_BASE + "varchar", "256"},
		{PROP_BASE + "bytearray", "1999"},
		{PROP_BASE + "jorm_chararray", "1999"},
		{PROP_BASE + "varchar.pk", "256"}
	};

	public static final Map properties = new HashMap(PROPS.length);

	static {
		for (int i = 0; i < PROPS.length; i++) {
			properties.put(PROPS[i][0],
					System.getProperty(PROPS[i][0], PROPS[i][1]));
		}
	}

	private final static int getDefaultSize(int idx) {
		return Integer.parseInt((String) properties.get(PROPS[idx][0]));
	}

	private int varcharSize = getDefaultSize(0);
	private int defaultBytearraySize = getDefaultSize(1);
	private int defaultChararraySize = getDefaultSize(2);
	private int varcharSizeInPk = getDefaultSize(3);

	public OracleAdapter() {
		super("oracle");
	}

	public OracleAdapter(String name) {
		super(name);
	}

	public void setVarcharSize(int s) {
		varcharSize = s;
	}

	public void setVarcharSizeInPk(int s) {
		varcharSizeInPk = s;
	}

	public boolean supportBatchPreparedStatement() {
		return false;
	}

	public String getSqlType(int typeCode, boolean usedInPK, int size, int scale) throws RdbAdapterException {
		int realsize = size;
		switch (typeCode) {
		case TYPECODE_BOOLEAN:
		case TYPECODE_OBJBOOLEAN:
		case TYPECODE_BYTE:
		case TYPECODE_OBJBYTE:
		case TYPECODE_SHORT:
		case TYPECODE_OBJSHORT:
		case TYPECODE_LONG:
		case TYPECODE_OBJLONG:
		case TYPECODE_DOUBLE:
		case TYPECODE_OBJDOUBLE:
		case TYPECODE_BIGINTEGER:
		case TYPECODE_BIGDECIMAL:
			return "NUMBER";
		case TYPECODE_STRING:
			if (realsize == NOSIZE) {
				realsize = (usedInPK ? varcharSizeInPk : varcharSize);
			}
			return "VARCHAR(" + realsize + ")";
		case TYPECODE_BYTEARRAY:
			if (realsize == NOSIZE) {
				realsize = defaultBytearraySize;
			}
			return (realsize > 2000 ? "LONG " : "") + "RAW(" + realsize + ")";
		case TYPECODE_CHARARRAY:
			if (realsize == NOSIZE) {
				realsize = defaultChararraySize;
			}
			return (realsize > 2000 ? "LONG " : "") + "RAW(" + realsize + ")";
		case TYPECODE_SERIALIZED:
			return "BLOB";
		case TYPECODE_DATE:
			return "DATE";
		default:
			return super.getSqlType(typeCode, usedInPK, size, scale);
		}
	}

	public String getSubstringName() {
		return ORACLESUBSTRING;
	}

	public String getFirstLocateExpression(String substring, String instring) {
		return ORACLEFIRSTLOCATE + "(" + instring + ", " + substring + ")";
	}

	public String getIndexedLocateExpression(String instring, String substring,
											 String fromIndex) {
		return ORACLEFIRSTLOCATE + "(" + instring + ", " + substring
				+ ", " + fromIndex + ")";
	}

	public void writeTableAlias(String alias, StringBuffer sb) {
		sb.append(" ");
		sb.append(alias);
	}
    public String getColumnAliasExpr(String aliasName) {
        return " \"" + aliasName +"\" ";
    }
    public void writeColumnAlias(String alias, StringBuffer sb) {
        sb.append(" \"" + alias + "\" ");
    }

	public String handleOrderBy(String query) {
		return "SELECT * FROM (" + query + ")";
	}

	public String getNextValInSequence(String seqName) {
		return "select " + seqName + " .nextval from dual";
	}

    // overrides existRelation for getTables with table names not supported by driver
    protected boolean existRelation(Connection connection, String relName, String[] relationTypes) throws SQLException {
        boolean existtable = false;
        ResultSet rs = null;
        
        if (relationTypes == null) {
            relationTypes = new String[]{"TABLE", "VIEW"};
        }
        
        try {
            rs = connection.getMetaData().getTables(null, null, null, relationTypes);
            while (rs.next() && !existtable) {
                existtable = relName.equalsIgnoreCase(rs.getString(3));
            }
        } finally {
            if (rs != null) {
                rs.close();
            }
        }
        return existtable;
    }

	public boolean existSequence(Connection connection, String seqName) throws SQLException {
		Statement s = connection.createStatement();
		s.execute(getCreateSequence(seqName));
		s.close();
		return true;
	}

	protected void modifyWhereClauseWithRange(StringBuffer sb,
											  String whereclause,
											  int rangeStart,
											  int rangeSize) {
		if (rangeSize < Integer.MAX_VALUE) {
			if (whereclause != null && whereclause.trim().length() > 0) {
				sb.append(" AND ");
			} else {
				sb.append(" WHERE ");
			}
			sb.append(" ROWNUM >" + rangeStart
					+ " AND ROWNUM <= " + rangeSize + rangeStart);
		}
	}

}
