/**
 * Copyright (C) 2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.jorm.mapper.rdb.adapter.api;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Connection;
import java.util.List;

/**
 * This interface permits to adapte the use of a JDBC driver and the SQL type
 * for the non conformance of a Database.
 * @author S.Chassande-Barrioz
 */
public interface RdbAdapter
	extends TypeAdapter,
	ResultsetAdapter,
	PreparedStatementAdapter,
	ValueAsSQLStringAdapter,
	SequenceAdapter {

	int NOSIZE = -1;


	/**
	 * @return The name of the RdbAdapter. This name is used in the factory to
	 * register the RdbAdapter instance. In addition it is supposes that the
	 * adapter is a part of the class name of the rdb adapter in respect to the
	 * following pattern: org.objectweb.jorm.mapper.rdb.XxxxAdapter where Xxxx
	 * is the adapter name with the upper first letter.
	 */
	String getName();

	/**
	 * Indicates if the driver supports the batch of PreparedStatement
	 */
	boolean supportBatchPreparedStatement();

	/**
	 * This method calculates the size of a result set window. If this feature is not
	 * supported by a database or its jdbc driver, this method returns -1.
	 * @param rs the result set which the size must be returned
	 */
	int fetchResultSetSize(ResultSet rs) throws SQLException;

	/**
	 * @return the name of the subString operator
	 */
	String getSubstringName();

	/**
	 * @return the use of the concat operator between two strings
	 */
	String getConcatExpression(String op1, String op2);

	/**
	 * Returns the expression for searching the position of the first occurrence
	 * of a substring in a string.
	 * <p>Two main syntaxes are found: position(substr in str) and
	 * instr(str, substr).
	 * @param substring The substring searched
	 * @param instring The string in which to search the substring
	 * @return The corresponding relational expression.
	 */
	String getFirstLocateExpression(String substring, String instring)
			throws RdbAdapterException;

	/**
	 * Returns the expression for searching the position of the first occurrence
	 * of a substring in a string starting from a given index.
	 * <p>Two main syntaxes are found: instr() and locate()
	 * @param inputString The string in which to search the substring
	 * @param substring The substring searched
	 * @param fromIndex The index from which to start searching
	 * @return The corresponding relational expression.
	 */
	String getIndexedLocateExpression(String inputString, String substring,
									  String fromIndex)
			throws RdbAdapterException;

	/**
	 * @return the name of the length operator
	 */
	String getLengthOperator() throws RdbAdapterException;

	/**
	 * Builds a query with join and table alias.
	 *
	 * The implementation must use writeTableAlias(String StringBuffer) method.
	 *
	 * @param selectClause is a string representation of the select clause
	 * @param jts is the tables with their join
	 * @param whereclause is a string representation of the where clause
	 * @param rangeFrom is the rank of the first row to be returned in case of
	 * a range limit, starting at 0
	 * @param rangeSize is the number of rows to be returned in case of a range
	 * limit. The convention is that Integer.MAX_VALUE means no range limit.
	 */
	String getQuery(String selectClause,
					List jts,
					String whereclause,
					int rangeFrom,
					int rangeSize);

	/**
     * Append a table alias to an SQL expression. To be used when constructing
     * the string expression of a from clause.
     * For example, most database recognize the syntax
     * 'select ... from R as ALIAS_FOR_R, S as ALIAS_FOR_S ... '
     * whereas Orabcle accepts the following syntax :
     * 'select ... from R ALIAS_FOR_R, S ALIAS_FOR_S ... '
	 * @param alias The table alias
	 * @param sb The StringBuffer into which to write the alias.
	 */
	void writeTableAlias(String alias, StringBuffer sb);

    /**
     * Append a column alias to an SQL expression. To be used when constructing
     * the string expression of a select clause.
     * For example, most database recognize the syntax
     * 'select A as ALIAS_FOR_A, B as ALIAS_FOR_B ... '
     * whereas Orabcle accepts the following syntax :
     * 'select A "ALIAS_FOR_A", B "ALIAS_FOR_B" '
     * @param alias the column alias
     * @param sb The StringBuffer into which to write the alias.
     */
    void writeColumnAlias(String alias, StringBuffer sb);

    /**
     * Returns a column alias expression. To be used when constructing
     * the string expression of a select clause.
     * For example, most database recognize the syntax
     * 'select A as ALIAS_FOR_A, B as ALIAS_FOR_B ... '
     * whereas Orabcle accepts the following syntax :
     * 'select A "ALIAS_FOR_A", B "ALIAS_FOR_B" '
     * @param aliasName the column alias
     */
    String getColumnAliasExpr(String aliasName);

	String handleOrderBy(String query);

	/**
	 * checks the table existence
	 * @param connection is the JDBC connection to use
	 * @param tableName is the name of the table
	 * @return true if the table exists, otherwise false.
	 */
	boolean existTable(Connection connection, String tableName)
			throws SQLException;

    /**
     * checks the view existence
     * @param connection is the JDBC connection to use
     * @param viewName is the name of the view
     * @return true if the view exists, otherwise false.
     */
    boolean existView(Connection connection, String viewName)
            throws SQLException;

    /**
	 * In the case the subquery needs to be aliased (Postgres), returns a string
	 * to be appended to the subquery.
	 * @return the String to be appended to the subquery
	 */
	String getSubqueryAlias();
}
